#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Jul  5 09:42:30 2020

@author: corkep
"""

from spatialmath import base
import matplotlib.pyplot as plt
import numpy as np

# geom2d
def plot_line():
    pass


    
    
def intersect_lines():
    # return mindist and points if no intersect
    pass

def distance_line_line():
    pass

def distance_line_point():
    pass

def points_join():

    pass

def intersect_polygon___line():
    pass

def contains_polygon_point():
    pass